#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## NACA 0012 wing configuration file for flow polar script
# Adrien Crovato

def getParam():
    p = {}
    # Specific
    span = 1. # span length
    # Input/Output
    p['File'] = '../models/n0012_3.geo' # Input file containing the model
    p['Pars'] = {'spn' : span, 'lgt' : 3., 'wdt' : 3., 'hgt' : 3., 'msLe' : 0.02, 'msTe' : 0.02, 'msF' : 1.} # Parameters for input file model
    p['Dim'] = 3 # Problem dimension
    p['Format'] = 'vtk' # Save format (vtk or gmsh)
    p['Slice'] = [0.01, 0.25, 0.5, 0.75, 0.95] # array of (y) coordinates to perform slice along the span (empty if none)
    p['TagId'] = 4 # tag number of physical group to be sliced
    # Markers
    p['Fluid'] = 'field' # Name of physical group containing the fluid
    p['Farfield'] = ['upstream', 'farfield', 'downstream'] # LIST of name of physical groups containing the farfield boundaries (upstream/downstream should be first/last element)
    p['Symmetry'] = 'symmetry' # Name of physical group containing the symmetry boundaries
    p['Wings'] = ['wing'] # LIST of names of physical groups containing the lifting surface boundary
    p['Wakes'] = ['wake'] # LIST of names of physical group containing the wake
    p['WakeTips'] = ['wakeTip'] # LIST of names of physical group containing the edge of the wake
    p['TeTips'] = ['teTip'] # LIST of names of physical group containing the edge of the wake and the trailing edge
    # Freestream
    p['M_inf'] = 0.3 # Freestream Mach number
    p['AoA_begin'] = -3. # Freestream angle of attack (begin)
    p['AoA_end'] = 3. # Freestream angle of attack (end)
    p['AoA_step'] = 1. # Freestream angle of attack (step)
    # Geometry
    p['S_ref'] = 1.*span # Reference surface length
    p['c_ref'] = 1. # Reference chord length
    p['x_ref'] = 0.25 # Reference point for moment computation (x)
    p['y_ref'] = 0. # Reference point for moment computation (y)
    p['z_ref'] = 0. # Reference point for moment computation (z)
    # Numerical
    p['LSolver'] = 'MUMPS' # Linear solver (Pardiso, GMRES, MUMPS or SparseLU)
    p['NSolver'] = 'Newton' # Noninear solver type (Picard or Newton)
    p['Rel_tol'] = 1e-6 # Relative tolerance on solver residual
    p['Abs_tol'] = 1e-8 # Absolute tolerance on solver residual
    p['Max_it'] = 10 # Solver maximum number of iterations
    p['LS_tol'] = 1e-6 # Tolerance on linesearch residual
    p['Max_it_LS'] = 10 # Linesearch maximum number of iterations
    p['AV_thrsh'] = 1e-2 # Residual threshold below which the artificial viscosity is decreased
    return p

def main():
    # Script call
    import flow.scripts.polar as p
    polar = p.Polar(getParam())
    polar.run()
    polar.disp()

if __name__ == "__main__":
    main()
