/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wHeatTerm.h"

#include "wElement.h"
#include "wCache.h"
#include "wGauss.h"
#include "wFct0.h"
#include "wFct2.h"

using namespace heat;
using namespace tbox;

/**
 * @brief Build volume terms for the heat equation on one element
 */
Eigen::MatrixXd HeatTerm::build(Element const &e, std::vector<double> const &u, Fct2 const &f, bool fake)
{
    // Get precomputed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();

    Eigen::MatrixXd K = Eigen::MatrixXd::Zero(e.nodes.size(), e.nodes.size());
    if (fake) // fake run - fill MPI job list
    {
        for (size_t k = 0; k < gauss.getN(); ++k)
        {
            Eigen::MatrixXd fk;
            f.eval(e, u, k, fk, fake);
        }
    }
    else // true run - use MPI cached results
    {
        // Gauss integration
        for (size_t k = 0; k < gauss.getN(); ++k)
        {
            // Jacobian inverse, shape functions and function evaluation
            Eigen::MatrixXd const &J = e.getJinv(k);
            Eigen::MatrixXd const &dff = cache.getDsf(k);
            Eigen::MatrixXd fk;
            f.eval(e, u, k, fk, fake);

            // Elementary stiffness matrix
            K += (fk * J * dff).transpose() * J * dff * gauss.getW(k) * e.getDetJ(k);
        }
    }
    return K;
}

/**
 * @brief Build surface/volume flux on one element
 */
Eigen::VectorXd HeatTerm::build(Element const &e, std::vector<double> const &u, Fct0 const &f)
{
    // Get precomputed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();

    // Gauss integration
    Eigen::VectorXd s = Eigen::VectorXd::Zero(e.nodes.size());
    for (size_t k = 0; k < gauss.getN(); ++k)
        s += cache.getSf(k) * f.eval(e, u, k) * gauss.getW(k) * e.getDetJ(k);
    return s;
}

/**
 * @brief Build volume flux of the heat equation on one element
 * @todo not used
 */
Eigen::VectorXd HeatTerm::build2(Element const &e, std::vector<double> const &u, Fct2 const &f)
{
    // Get precomputed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();

    // Gauss integration
    Eigen::VectorXd qV = Eigen::VectorXd::Zero(e.nodes.size());
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Shape functions, gradient and flux evaluation
        Eigen::MatrixXd const &dff = cache.getDsf(k);
        Eigen::MatrixXd fk(2, 2);
        f.eval(e, u, k, fk, false);

        // Elementary flux vector
        qV += (fk * e.computeGradient(u, k)).transpose() * e.getJinv(k) * dff * gauss.getW(k) * e.getDetJ(k);
    }
    return qV;
}

/**
 * @brief Compute (integrate) volume flux
 */
Eigen::VectorXd HeatTerm::computeFlux(Element const &e, std::vector<double> const &u, Fct2 const &f)
{
    // Get precomputed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();

    // Gauss integration
    Eigen::VectorXd qV = Eigen::Vector2d::Zero(); // 2 dimensions
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Gradient and matrix function evaluation
        Eigen::VectorXd gradk = e.computeGradient(u, k);
        Eigen::MatrixXd fk(2, 2);
        f.eval(e, u, k, fk, false);

        qV -= fk * gradk * gauss.getW(k) * e.getDetJ(k);
    }
    return qV;
}

/**
 * @brief Compute (integrate) matrix
 */
Eigen::MatrixXd HeatTerm::computeMatrix(Element const &e, std::vector<double> const &u, Fct2 const &f)
{
    // Get precomputed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();

    // Gauss integration
    Eigen::MatrixXd out = Eigen::Matrix2d::Zero(); // 2 dimensions, should match fk size
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Function evaluation
        Eigen::MatrixXd fk;
        f.eval(e, u, k, fk, false);

        out += fk * gauss.getW(k) * e.getDetJ(k);
    }
    return out;
}
