/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wQuad4.h"
#include "wSfQuad4.h"
#include "wGaussQuad4.h"
#include "wCacheQuad4.h"
#include "wNode.h"

using namespace tbox;

Quad4::Quad4(int n, Tag *_ptag, Tag *_etag,
             std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
}

/**
 * @brief Return cache holding shape functions and Gauss points for all Quad4 elements
 */
Cache &Quad4::getVCache() const
{
    return getCache(order);
}

/**
 * @brief Initialize precomputed values
 */
void Quad4::initValues(bool isvol, int ordr)
{
    // Set flags
    isVol = isvol;
    order = ordr;
    hasVals = true;
    // Update values
    size_t ngp = getCache(order).getVGauss().getN();
    Js.resize(ngp);
    iJs.resize(ngp);
    detJs.resize(ngp);
    this->update();
}
/**
 * @brief Initialize precomputed gradients
 */
void Quad4::initGradients()
{
    // Set flags
    hasGrads = true;
    // Update values
    size_t ngp = getCache(order).getVGauss().getN();
    gradJs.resize(ngp);
    gradDetJs.resize(ngp);
    this->update();
}
/**
 * @brief Update precomputed values and gradients
 */
void Quad4::update()
{
    // Update values
    if (hasVals)
    {
        // Jacobian and normal
        // Quad4 is not of same dimension as Problem (surface element)
        if (!isVol)
        {
            this->buildSurfaceJ();
            this->computeNormal();
        }
        // Quad4 is of same dimension as Problem (volume element)
        else
            this->buildJ();
        // Volume and CG
        this->computeV();
        this->computeCg();
    }
    // Update gradients
    if (hasGrads)
    {
        // Jacobian, volume and normal
        // Quad4 is not of same dimension as Problem (surface element)
        if (!isVol)
        {
            this->buildGradientSurfaceJ();
            this->buildGradientV(3); // gradient of surface
            this->computeGradientNormal();
        }
        // Quad4 is of same dimension as Problem (volume element)
        else
        {
            this->buildGradientJ();
            this->buildGradientV(2); // gradient of volume
        }
    }
}

/**
 * @brief Compute Jacobian determinant
 *
 * detJ = norm(x_i dN_i,xi X x_i dN_i,eta)
 */
void Quad4::buildSurfaceJ()
{
    for (size_t k = 0; k < getCache(order).getVGauss().getN(); ++k)
    {
        std::vector<Eigen::Vector3d> ts = this->buildTangents(k);
        detJs[k] = ts[0].cross(ts[1]).norm();
    }
}
/**
 * @brief Compute Jacobian matrix, inverse and determinant at Gauss point k
 *
 * J_ij(2,2) = diN_n * xj_n (n is the node)
 */
void Quad4::buildJ()
{
    CacheQuad4 &cache = getCache(order);
    for (size_t k = 0; k < cache.getVGauss().getN(); ++k)
    {
        Eigen::MatrixXd const &dff = cache.getDsf(k);

        Eigen::Matrix2d JJ = Eigen::Matrix2d::Zero(); // temporary fixed-size matrix to efficiently compute the inverse
        size_t i = 0;
        for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
            JJ += dff.col(i) * (*it)->pos.transpose().block<1, 2>(0, 0);
        Js[k] = JJ;
        iJs[k] = JJ.inverse();
        detJs[k] = JJ.determinant();
        this->checkJac(k);
    }
}

/**
 * @brief Return the element surface
 */
void Quad4::computeV()
{
    CacheQuad4 &cache = getCache(order);
    GaussQuad4 &gauss = cache.gauss;

    // Gauss integration
    vol = 0.0;
    for (size_t k = 0; k < gauss.getN(); ++k)
        vol += gauss.getW(k) * this->getDetJ(k);
}

/**
 * @brief Compute element tangent vectors
 */
std::vector<Eigen::Vector3d> Quad4::computeTangents() const
{
    Eigen::Vector3d x1 = nodes[0]->pos;
    Eigen::Vector3d x3 = nodes[2]->pos;
    Eigen::Vector3d x4 = nodes[3]->pos;

    return std::vector<Eigen::Vector3d>{(x3 - x1), (x4 - x1)};
}

/**
 * @brief Build element tangent vectors using shape functions at Gauss point k
 */
std::vector<Eigen::Vector3d> Quad4::buildTangents(size_t k) const
{
    Eigen::MatrixXd const &dff = getCache(order).getDsf(k);

    std::vector<Eigen::Vector3d> t(2, Eigen::Vector3d::Zero());
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        t[0] += (*it)->pos * dff(0, i);
        t[1] += (*it)->pos * dff(1, i);
    }
    return t;
}

/**
 * @brief Compute element unit normal vector
 */
void Quad4::computeNormal()
{
    std::vector<Eigen::Vector3d> ts = this->computeTangents();
    normal = ts[0].cross(ts[1]).normalized();
}
